import wrappy
import robin_stocks.robinhood as rh
from tr4d3r.core.asset import Cash
from tr4d3r.core.portfolio import RealTimePortfolio
from .asset import RobinhoodRealItem
from .market import RobinhoodRealMarket
from .utils import cached_order_info
from collections import defaultdict
from pprint import pformat


class RobinhoodRealPortfolio(RealTimePortfolio):
    """
    A real portfolio on Robinhood.
    """
    
    DEFAULT_CASH_CLASS = Cash
    CASH_NAME = 'USD'
    DEFAULT_ITEM_CLASS = RobinhoodRealItem
    DUMMY_ITEM = RobinhoodRealItem('Robinhood-N/A', quantity=0.0, unit_cost=0.0)
    
    def __init__(self, principal, cash, items, open_orders, **kwargs):
        """
        Log -> Load -> Check -> Construct.
        """
        #self._info("Logging in to Robinhood..")
        #rh.login()
        
        # constructor
        self._info("Constructing portfolio..")
        super().__init__(principal, cash, items, open_orders, **kwargs)
        self.market = RobinhoodRealMarket()
       
        # resolve filled orders, consistency check
        # get information from Robinhood
        self._info("Checking online info..")
        self.refresh_open_orders()
        self.validate_cash()
        self.validate_items()
        self._good(f"Checks passed.\n{pformat(self.to_dict())}")
        
    @classmethod
    def order_brief(cls, order_dict, detail_level=0):
        """
        Get the most critical information of an order.
        """
        if order_dict is None:
            return dict(info='No transaction')
        
        assert isinstance(detail_level, int)
        symbol = rh.get_instrument_by_url(order_dict['instrument'], info='symbol')
        amount = float(order_dict['total_notional']['amount'])
        brief = dict(symbol=symbol)        
        brief.update({_k: order_dict.get(_k, None) for _k in ['type', 'side', 'state', 'price', 'quantity']})
        brief.update(dict(amount=amount))
        if detail_level >= 1:
            brief.update({_k: order_dict.get(_k, None) for _k in ['id', 'created_at']})
        return brief
    
    @classmethod
    def order_side_to_bool(cls, side):
        if side == 'buy':
            return True
        elif side == 'sell':
            return False
        else:
            raise ValueError(f"Expected buy or sell, got {side}")
        
    def open_order_values(self):
        value_dict = defaultdict(float)
        for _order_id in self.open_orders:
            _order_dict = cached_order_info(_order_id)
            _symbol, _side, _amount = _order_dict['symbol'], _order_dict['side'], float(_order_dict['total_notional']['amount'])
            
            _buy_flag = self.__class__.order_side_to_bool(_side)
            _coeff = 1 if _buy_flag else -1
            value_dict[_symbol] += _coeff * _amount
        return value_dict
            
    def resolve_filled_order(self, order):
        """
        Expected to be executed exactly once for each order.
        """
        side = order['side']
        symbol = order['symbol']
        buy_flag = self.__class__.order_side_to_bool(side)
        
        assert len(order['executions']) == 1, f"Expected exactly one execution, got {order['executions']}"
        unit_cost = float(order['executions'][0]['price'])
        quantity = float(order['executions'][0]['quantity'])
        amount = float(order['executed_notional']['amount'])
        currency = order['executed_notional']['currency_code']
        cash = self.__class__.DEFAULT_CASH_CLASS(currency, amount)
        item = self.__class__.DEFAULT_ITEM_CLASS(symbol, quantity, unit_cost)
        
        if buy_flag:
            self.cash.sub(cash)
            self.add_item(item)
        else:
            self.sub_item(item)
            self.cash.add(cash)
            self.check(item.name)
            
    def refresh_single_open_order(self, order_id):
        """
        Refresh the state of an open order.
        If it is now filled, resolve its transaction and close it.
        If it has failed / been cancelled, make an alert.
        """
        order = rh.get_stock_order_info(order_id)
        assert order['type'] == 'market', f"Expected a market order, got {order['type']}"
            
        brief = self.__class__.order_brief(order)
        order['symbol'] = brief['symbol']
        
        state = order['state']
        if state in {'unconfirmed', 'confirmed', 'queued'}:
            # stay in the open order list
            self._info(f"Order pending: {brief}")
        elif state == 'filled':
            # read quantity/amount and change cash/item
            self._good(f"Order complete: {brief}")
            self.open_orders.remove(order_id)
            self.resolve_filled_order(order)
        elif state in {'failed', 'canceled'}:
            self._fail(f"Order failed: {brief}")
            self.open_orders.remove(order_id)
        else:
            raise ValueError(f"Unexpected state {state} for order id {order_id}")
        
    def refresh_open_orders(self):
        """
        Refresh the state of all open orders.
        """
        open_order_ids = self.open_orders.copy()
        for _order_id in open_order_ids:
            self.refresh_single_open_order(_order_id)
    
    def validate_cash(self):
        """
        Subroutine of the constructor. Validate local cash against online cash.
        """
        cash = self.cash
        available_cash = float(rh.load_account_profile('cash'))
        assert cash.name == self.__class__.CASH_NAME, f"Unexpected cash name: {cash.name}"
        assert cash.quantity <= available_cash, f"Expected cash >= {cash.quantity}, got {available_cash}"
        
    def validate_items(self):
        """
        Subroutine of the constructor. Validate local items against online holdings.
        """
        holdings_dict = rh.build_holdings()
        for _item in self.item_dict.values():
            assert _item.name in holdings_dict, f"Expected item {_symbol} in holdings"
            _online_dict = holdings_dict[_item.name]
            
            _offline_q = _item.quantity
            _online_q = float(_online_dict["quantity"])
            assert _offline_q == _online_q, f"Expected {_item.name} quantity {_offline_q}, got {_online_q}"
            
            _offline_c = _item.unit_cost
            _online_c = float(_online_dict["average_buy_price"])
            assert _offline_q == _online_q, f"Expected {_item.name} unit cost {_offline_c}, got {_online_c}"
    
    def market_buy(self, *args, **kwargs):
        """
        Customizing over parent method: keep a brief version of info.
        """
        order = super().market_buy(*args, **kwargs)
        return self.__class__.order_brief(order)
    
    def market_sell(self, *args, **kwargs):
        """
        Customizing over parent method: keep a brief version of info.
        """
        order = super().market_sell(*args, **kwargs)
        return self.__class__.order_brief(order)
    
    def local_deposit(self, amount):
        """
        Put more cash under this particular portfolio.
        """
        if amount == 0:
            self._warn("Trivial deposit amount")
            return
        assert amount > 0, "Invalid deposit amount: {amount}"
        # make sure there is enough cash
        target = self.cash.quantity + amount
        available_cash = float(rh.load_account_profile('cash'))
        assert target <= available_cash, f"Expected cash >= {target}, got {available_cash}"
        self.principal.quantity += amount
        self.cash.quantity = target
        
        return
    
    def local_withdraw(self, amount):
        """
        Pull cash from this particular portfolio.
        """
        if amount == 0:
            self._warn("Trivial withdrawal amount")
            return
        assert amount > 0, "Invalid withdrawal amount: {amount}"
        assert self.cash.quantity >= amount, f"Not enough cash ({self.cash.quantity}) to withdraw ({amount})"
        self.principal.quantity -= amount
        self.cash.quantity -= amount
        
        return


"""
Specialized assets in the Robinhood scenario.
"""
from tr4d3r.core.asset import Item
from tr4d3r.utils.misc import timed_lru_cache
from .utils import symbol_to_market_mic
import robin_stocks.robinhood as rh


class RobinhoodRealItem(Item):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.market_mic = symbol_to_market_mic(self.name)
    
    @classmethod
    def from_rh_holding(cls, symbol, data_dict):
        quantity = data_dict["quantity"]
        unit_cost = data_dict["average_buy_price"]
        return cls(symbol, quantity, unit_cost)
    
    @timed_lru_cache(seconds=60)
    def unit_price(self):
        price_list = rh.get_latest_price(self.name, includeExtendedHours=True)
        return list(map(float, price_list))[0]
    
    @timed_lru_cache(seconds=60)
    def bid_price(self):
        price_list = rh.get_latest_price(self.name, 'bid_price')
        return list(map(float, price_list))[0]
    
    @timed_lru_cache(seconds=60)
    def ask_price(self):
        price_list = rh.get_latest_price(self.name, 'ask_price')
        return list(map(float, price_list))[0]
    
    


from tr4d3r.core import Loggable
from tr4d3r.core.asset import Item
from tr4d3r.core.mechanism import NamedCallback, Tentative
from tr4d3r.utils.misc import datetime_to_string, string_to_datetime, utcnow
from pprint import pformat
from datetime import datetime
from collections import defaultdict
from abc import abstractmethod
import wrappy
import dill as pickle


class EquilibriumPortfolioManager(Loggable):
    """
    Manages any number of portfolios with
        - a fixed set of items considered for trading;
        - an adjustable target ratio in terms of total worth, for each item.
    """
    # daily progression of 3%, capped at 50% for any time gap
    DEFAULT_PROGRESSION_FUNC = lambda t: min(0.5, 0.03 * (t / 86400))
    
    def __init__(self, equilibrium, params=None):
        """
        equilibrium: symbol -> target_ratio dict.
        """
        self._initial_equilibrium = equilibrium.copy()
        self._running_equilibrium = equilibrium.copy()
        self.params = params or {}
    
    def __repr__(self):
        return self.params.get('name', self.__class__.__name__)
    
    @property
    def initial_equilibrium(self):
        return self._initial_equilibrium
    
    @property
    def equilibrium(self):
        return self._running_equilibrium
    
    @equilibrium.setter
    def equilibrium(self, equil_dict):
        self._info(f"Setting new equilibrium {equil_dict}")
        assert sum(equil_dict.values()) <= 1.0 - self.params.get('cash_ratio', 0.03), "Too much total ratio for items"
        self._running_equilibrium = equil_dict.copy()
    
    def linear_update(self, equil_dictl, coefficients):
        """
        Update the running equilibrium as a linear combination of equilibria.
        """
        # initialize unnormalized linear combination
        raw_combination = defaultdict(float)
        total_coeff = 0.0
        
        # collect weighted ratios for each symbol
        for _dict, _coeff in zip(equil_dictl, coefficients):
            # the total ratio supplied by each equil_dict must not exceed 1.0-cash
            _total_ratio = 0.0
            for _symbol, _ratio in _dict.items():
                assert _ratio >= 0.0, f"Expected non-negative ratio, got {_ratio}"
                if _ratio > 0.0:
                    raw_combination[_symbol] += _ratio * _coeff
                    _total_ratio += _ratio
            assert _total_ratio <= 1.0 - self.params.get('cash_ratio', 0.03), "Too much total ratio for items"
            total_coeff += _coeff
        
        # normalize and set as new equilibrium
        new_equilibrium = {_k: _v / total_coeff for _k, _v in raw_combination.items()}
        self.equilibrium = new_equilibrium
        
    @abstractmethod
    def tick_read(self, folio, **kwargs):
        pass
    
    @abstractmethod
    def tick_write(self, folio, **kwargs):
        pass
    
    @classmethod
    def load_pickle(cls, pkl_path):
        with open(pkl_path, 'rb') as f:
            pkl_dict = pickle.load(f)
        equil_dict = pkl_dict['equilibrium']
        param_dict = pkl_dict['params']
        return cls(equil_dict, param_dict)
    
    def dump_pickle(self, pkl_path):
        pkl_dict = {
            'equilibrium': self.equilibrium,
            'params': self.params,
        }
        with open(pkl_path, 'wb') as f:
            pickle.dump(pkl_dict, f)
    
    
class PseudoTimeEquilibrium(EquilibriumPortfolioManager):
    def tick_read(self, folio, time):
        """
        Gather data for analysis.
        """
        worth = folio.worth(time)
        data = {
            'time': time,
            'principal': folio.principal.quantity,
            'worth': worth,
            '% cash': 100 * folio.cash.worth() / worth,
        }
        for _symbol in self.equilibrium.keys():
            data[f"P:{_symbol}"] = folio.market.get_price(_symbol, time)
            data[f"Q:{_symbol}"] = folio[_symbol].quantity
            data[f"C:{_symbol}"] = folio[_symbol].unit_cost
            data[f"% {_symbol}"] = 100 * folio[_symbol].worth(time) / worth
        
        return data
        
    def tick_write(self, folio, time):
        """
        Make trading decisions and update the last time of making moves.
        """
        gap_seconds = folio.tick(time)
        for _symbol, _ratio in self.equilibrium.items():
            # determine if item exists and then its worth
            _item = folio[_symbol]
            _item_exists = _item.name == _symbol
            _bid_worth = _item.bid_worth(time) if _item_exists else 0.0
            _ask_worth = _item.ask_worth(time) if _item_exists else 0.0
            
            # determine "step size"
            _target_worth = _ratio * folio.worth(time)
            _step = self.params.get('progression_func', self.__class__.DEFAULT_PROGRESSION_FUNC)(gap_seconds)
            assert _step <= 1.0, f"Step size too large: {_step}"
            
            # market orders toward equilibrium
            # ask_worth is always above bid_worth
            if _target_worth > _ask_worth:
                _amount = (_target_worth - _ask_worth) * _step
                folio.market_buy(_symbol, amount=_amount, time=time)
            elif _item_exists and _target_worth < _bid_worth:
                _amount = (_bid_worth - _target_worth) * _step
                folio.market_sell(_symbol, amount=_amount, time=time)
            else:
                pass
            

class RealTimeEquilibrium(EquilibriumPortfolioManager):
    def tick_read(self, folio):
        """
        Gather data for analysis.
        """
        worth = folio.worth()
        data = {
            'time': utcnow(),
            'principal': folio.principal.quantity,
            'worth': worth,
            '% cash': 100 * folio.cash.worth() / worth,
        }
        for _symbol in self.equilibrium.keys():
            _price = folio.market.get_price(_symbol)
            _quantity = folio[_symbol].quantity
            data[f"P:{_symbol}"] = _price
            data[f"Q:{_symbol}"] = _quantity
            data[f"C:{_symbol}"] = folio[_symbol].unit_cost
            data[f"% {_symbol}"] = 100 * (_price * _quantity) / worth
        
        return data
        
    def tick_write(self, folio, execute=False):
        """
        Make trading decisions and update the last time of making moves.
        """
        prev_datetime = datetime_to_string(folio.last_tick_time)
        gap_seconds = folio.tick(update=execute)
        self._info(f"Gap {gap_seconds} seconds since {prev_datetime}")
        folio_worth = folio.worth()
        open_order_values = folio.open_order_values()
        for _symbol, _ratio in self.equilibrium.items():
            # determine if item exists and then its worth
            _item = folio[_symbol]
            _item_exists = _item.name == _symbol
            _ava_worth = _item.worth() if _item_exists else 0.0
            _bid_worth = _item.bid_worth() if _item_exists else 0.0
            _ask_worth = _item.ask_worth() if _item_exists else 0.0
            # adjust estimated worth based on open order
            _ord_worth = open_order_values[_symbol]
            _cur_worth = _ava_worth + _ord_worth
            _bid_worth += _ord_worth
            _ask_worth += _ord_worth
            
            # determine "step size"
            _cur_ratio = _cur_worth / folio_worth
            _target_worth = _ratio * folio_worth
            _step = self.params.get('progression_func', self.__class__.DEFAULT_PROGRESSION_FUNC)(gap_seconds)
            assert _step <= 1.0, f"Step size too large: {_step}"
            
            # market orders toward equilibrium
            # ask_worth is always above bid_worth
            self._info(f"{_symbol} worth : tar. {round(_target_worth, 2)} ({round(_ratio*100, 2)}%) | cur. {round(_cur_worth, 2)} ({round(_cur_ratio*100, 2)}%)")
            self._info(f"{_symbol} worth : ava. {round(_ava_worth, 6)} | ord. {round(_ord_worth, 2)} | bid-ask {round(_bid_worth, 2)}-{round(_ask_worth, 2)}")
            if _target_worth > _ask_worth:
                _amount = (_target_worth - _cur_worth) * _step
                _action = 'market buy'
                _func = folio.market_buy
            elif _item_exists and _target_worth < _bid_worth:
                _amount = (_cur_worth - _target_worth) * _step
                _action = 'market sell'
                _func = folio.market_sell
            else:
                self._info(f"{_symbol} near equilibrium, staying put.")
                return gap_seconds
            
            _base_msg = f"{_action}: {_symbol} | {round(_amount, 6)} {folio.cash.name}"
            if execute:
                info = _func(_symbol, amount=_amount)
                self._warn(f"Real {_base_msg}\n{info}")
            else:
                self._info(f"Fake {_base_msg}")
            
        return gap_seconds


import wrappy
import numpy as np
import pandas as pd
import robin_stocks.robinhood as rh
import math
from tqdm import tqdm
from datetime import datetime, timezone, timedelta
from pprint import pformat
from tr4d3r.core.market import PseudoTimeMarket, RealTimeMarket
from tr4d3r.utils.misc import autodatetime


@wrappy.memoize(cache_limit=100, return_copy=False, persist_path='robin-history.pkl', persist_batch_size=1)
def rh_historicals(symbol, interval, span):
    """
    Helper function for the RobinhoodPseudoTime class.
    """
    rh.login()
    dictl = rh.get_stock_historicals(symbol, interval=interval, span=span)
    assert isinstance(dictl, list) and dictl, f"Expected non-empty list of dicts, got {dictl}"
    assert isinstance(dictl[0], dict), f"Expected non-empty list of dicts, got {dictl}"
    return dictl

def rh_str_to_datetime(utc_str):
    raw_dt = datetime.strptime(utc_str, '%Y-%m-%dT%H:%M:%SZ')
    utc_dt = raw_dt.astimezone(timezone.utc)
    return utc_dt

class RobinhoodPseudoMarket(PseudoTimeMarket):
    """
    Robinhood pseudo-time market where orders can take place in the past.
    This is for simulation purposes.
    """
    CURRENCY = 'USD'
    
    def __init__(self, symbols, interval='day', span='5year'):
        assert len(symbols) > 0, "Expecting at least one symbol"
        self.span = span
        self.interval = interval
        
        self.symbol_to_table = {}
        for _symbol in tqdm(symbols, desc="Loading stock historicals"):
            _dictl = rh_historicals(_symbol, interval, span)
            _table = pd.DataFrame(_dictl)
            _table['begins_at'] = _table['begins_at'].apply(rh_str_to_datetime)
            _table['begins_at_timestamp'] = _table['begins_at'].apply(lambda x: x.timestamp())
            self.symbol_to_table[_symbol] = _table
        
        first_table = self.symbol_to_table[symbols[0]]
        last_idx = first_table.shape[0] - 1
        self.start_time = first_table.loc[0, 'begins_at']
        self.end_time = first_table.loc[last_idx, 'begins_at']
        
    def __repr__(self):
        info = {
            'listing': list(self.symbol_to_table.keys()),
            'timespan': self.span,
            'interval': self.interval,
        }
        return f"===={self.__class__.__name__}====\n{pformat(info)}"
    
    def time_sequence(self, offset):
        offset_list = super().time_sequence(offset)
        return [self.start_time + _off for _off in offset_list]
        
    def get_price(self, symbol, time=None):
        table = self.symbol_to_table[symbol]
        time = autodatetime(time)
        
        # find the most recent recorded time before the supplied time
        interval_idx = np.searchsorted(table['begins_at_timestamp'].values, time.timestamp()+1e-6) - 1
        assert interval_idx >= 0, f"Time out of range: {time}"
        
        return float(table.loc[interval_idx, 'open_price'])

    
class RobinhoodRealMarket(RealTimeMarket):
    """
    Robinhood real-time market.
    """
    MIN_TRANS_AMOUNT = 1e-2
    MIN_TRANS_SHARES = 1e-6
    
    def __init__(self):
        pass
    
    @classmethod
    def round_shares(cls, shares):
        return math.floor(shares / cls.MIN_TRANS_SHARES) * cls.MIN_TRANS_SHARES
        
    @classmethod
    def round_amount(cls, amount):
        return math.floor(amount / cls.MIN_TRANS_AMOUNT) * cls.MIN_TRANS_AMOUNT
        
    def validate_order(self, symbol, shares, amount):
        # check symbol exists in the market
        price = self.get_price(symbol)
        
        # consistency between shares and amount
        if shares is not None:
            assert amount is None, "Expected either shares or amount, got both"
            assert isinstance(shares, float), f"Invalid shares: {shares}"
            amount = shares * price
        else:
            assert amount is not None, "Expected either shares or amount, got neither"
        
        amount = self.__class__.round_amount(amount)
        shares = self.__class__.round_shares(amount / price)
        
        if shares < self.__class__.MIN_TRANS_SHARES or amount < self.__class__.MIN_TRANS_AMOUNT:
            return dict(symbol=symbol, shares=0.0, amount=None, valid=False)
        else:
            return dict(symbol=symbol, shares=shares, amount=None, valid=True)
            
    def market_buy(self, symbol, shares=None, amount=None, **kwargs):
        validate_dict = self.validate_order(symbol, shares, amount)
        if validate_dict['valid']:
            order = rh.orders.order_buy_fractional_by_quantity(symbol, validate_dict['shares'], **kwargs)
        else:
            order = None
        return order
    
    def market_sell(self, symbol, shares=None, amount=None, **kwargs):
        validate_dict = self.validate_order(symbol, shares, amount)
        if validate_dict['valid']:
            order = rh.orders.order_sell_fractional_by_quantity(symbol, validate_dict['shares'], **kwargs)
        else:
            order = None
        return order
    
    def get_price(self, symbol):
        price_list = rh.stocks.get_latest_price(symbol, includeExtendedHours=False)
        return list(map(float, price_list))[0]
    
    def get_bid_price(self, symbol):
        price_list = rh.stocks.get_latest_price(symbol, 'bid_price', includeExtendedHours=False)
        return list(map(float, price_list))[0]
    
    def get_ask_price(self, symbol, time=None):
        price_list = rh.stocks.get_latest_price(symbol, 'ask_price', includeExtendedHours=False)
        return list(map(float, price_list))[0]
    


from rich.console import Console
from rich.prompt import Prompt, Confirm

class Loggable:
    """
    Base class that provides consistently templated logging.

    Inspired by `wasabi`'s `good`/`info`/`warn`/`fail` methods.

    [`Rich` style guide](https://rich.readthedocs.io/en/latest/style.html)
    """

    CONSOLE = Console()

    def _print(self, *args, **kwargs):
        self.__class__._print(*args, **kwargs)

    def _good(self, message):
        self.__class__._good(message)

    def _info(self, message):
        self.__class__._info(message)

    def _warn(self, message):
        self.__class__._warn(message)

    def _fail(self, message):
        self.__class__._fail(message)
    
    @classmethod
    def _print(cls, *args, **kwargs):
        cls.CONSOLE.print(*args, **kwargs)

    @classmethod
    def _prompt(cls, message, **kwargs):
        answer = Prompt.ask(
            f":question: {cls.__name__}: {message}",
            **kwargs,
        )
        return answer

    @classmethod
    def _confirm(cls, message, **kwargs):
        answer = Confirm.ask(
            f":question: {cls.__name__}: {message}",
            **kwargs,
        )
        return answer

    @classmethod
    def _good(cls, message):
        cls.CONSOLE.print(
            f":green_circle: {cls.__name__}: {message}",
            style="green",
        )

    @classmethod
    def _info(cls, message):
        cls.CONSOLE.print(
            f":blue_circle: {cls.__name__}: {message}", style="blue"
        )

    @classmethod
    def _warn(cls, message):
        cls.CONSOLE.print(
            f":yellow_circle: {cls.__name__}: {message}",
            style="yellow",
        )

    @classmethod
    def _fail(cls, message):
        cls.CONSOLE.print(
            f":red_circle: {cls.__name__}: {message}", style="red"
        )
    
        


import re
import robin_stocks.robinhood as rh
from functools import lru_cache


def symbol_to_market_mic(symbol):
    """
    Get the MIC value of the market for a specified symbol.
    """
    markets = rh.get_instruments_by_symbols(symbol, info='market')
    
    if not markets:
        return 'N/A'
    
    assert len(markets) == 1, f"Expecting exactly one market for symbol {symbol}, got {markets}"
    mic = re.search(r'[A-Z]{4}\/$', markets[0]).group()[:-1]
    return mic


@lru_cache(maxsize=1000)
def cached_order_info(order_id):
    """
    A cached version of get_stock_order_info.
    Can be useful for quickly finding invariant attributes
    such as the trade amount or quantity, but not the state.
    """
    order_dict = rh.get_stock_order_info(order_id)
    symbol = rh.get_instrument_by_url(order_dict['instrument'], info='symbol')
    order_dict['symbol'] = symbol
    return order_dict


from datetime import datetime, timedelta, timezone
from functools import lru_cache, wraps

DATETIME_FORMAT = "%Y%m%d %H:%M:%S.%f %z"

def utcnow():
    return datetime.now().astimezone(timezone.utc)

def datetime_to_string(dt):
    assert isinstance(dt, datetime), f"Expected datetime, got {type(dt)} {dt}"
    dt_utc = dt.astimezone(timezone.utc)
    dt_str = dt_utc.strftime(DATETIME_FORMAT)
    return dt_str

def string_to_datetime(dt_str):
    assert isinstance(dt_str, str), f"Expected string, got {type(dt_str)} {dt_str}"
    dt = datetime.strptime(dt_str, DATETIME_FORMAT)
    dt_utc = dt.astimezone(timezone.utc)
    return dt_utc
    
def autostamp(time):
    time = time or utcnow()
    if isinstance(time, datetime):
        time = time.timestamp()
    assert isinstance(time, float), f"Expected float, got {type(time)} {time}"
    return time

def autodatetime(dt):
    dt = dt or utcnow()
    if isinstance(dt, float):
        dt = datetime.fromtimestamp(dt).astimezone(timezone.utc)
    assert isinstance(dt, datetime), f"Expected datetime, got {type(dt)} {dt}"
    return dt
    
def timed_lru_cache(seconds: int, maxsize: int = 128):
    def wrapper_cache(func):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = utcnow() + func.lifetime

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            current_time = utcnow()
            if current_time >= func.expiration:
                func.cache_clear()
                func.expiration = current_time + func.lifetime

            return func(*args, **kwargs)

        return wrapped_func

    return wrapper_cache


import math
from numpy.random import default_rng
from datetime import datetime, timedelta, timezone
from pprint import pformat
from tr4d3r.core.market import PseudoTimeMarket
from tr4d3r.utils.misc import autodatetime
from tr4d3r.utils.random import exponential_random_walk


class DummyPseudoTimeMarket(PseudoTimeMarket):
    """A fake market of exponential random walks.
    """
    START_TIME = datetime.strptime('20160101 00:00:00', '%Y%m%d %H:%M:%S').astimezone(timezone.utc)
    END_TIME = datetime.strptime('20220101 00:00:00', '%Y%m%d %H:%M:%S').astimezone(timezone.utc)
    TICK_PERIOD = 10
    STEPS = math.floor((END_TIME - START_TIME).total_seconds() / TICK_PERIOD) - 1
    
    INITIAL_LISTING = {
        'Gold': {'price': 100.0, 'volatility': 1e-5},
        'Stock-1': {'price': 10.0, 'volatility': 3e-5},
        'Stock-2': {'price': 10.0, 'volatility': 3e-5},
        'Coin-1': {'price': 1.0, 'volatility': 1e-4},
        'Coin-2': {'price': 1.0, 'volatility': 1e-4},
        'Coin-3': {'price': 1.0, 'volatility': 1e-4},
    }
    CURRENCY = 'USD'
    
    def __init__(self, seed=0):
        self.rng = default_rng(seed=seed)
        steps = self.__class__.STEPS
        self.price_history = {
            _name: exponential_random_walk(
                self.rng, 
                _attr['price'], 
                self.__class__.STEPS, 
                loc=0.0, 
                scale=_attr['volatility'] * self.__class__.TICK_PERIOD,
            ) for _name, _attr in self.__class__.INITIAL_LISTING.items()
        }
   
    def __repr__(self):
        info = {
            'listing': list(self.price_history.keys()),
            'timespan': [
                self.__class__.START_TIME.strftime('%Y%m%d %H:%M'),
                self.__class__.END_TIME.strftime('%Y%m%d %H:%M'),
            ],
            'tick_period': self.__class__.TICK_PERIOD,
        }
        return f"===={self.__class__.__name__}====\n{pformat(info)}"
    
    def time_sequence(self, offset):
        offset_list = super().time_sequence(offset)
        return [self.__class__.START_TIME + _off for _off in offset_list]
        
    def get_price(self, symbol, time=None):
        time = autodatetime(time)
        past_seconds = (time - self.__class__.START_TIME).total_seconds()
        index = math.floor(past_seconds / self.__class__.TICK_PERIOD)
        assert symbol in self.price_history, f"Unrecognized item {symbol}"
        history = self.price_history[symbol]
        return history[min(index, len(history)-1)]
    


"""
Asset components.
"""
from abc import abstractmethod
import pandas as pd
from tr4d3r.core import Loggable
from tr4d3r.utils.misc import autostamp
from datetime import timedelta


class PseudoTimeMarket(Loggable):
    """Abstract base class of markets where one can buy and sell items.
    """
    
    def __init__(self, *args, **kwargs):
        pass
    
    @abstractmethod
    def time_sequence(self, offset):
        pass
    
    @abstractmethod
    def get_price(self, symbol):
        pass
    
    def create_portfolio(self, cash_quantity, **kwargs):
        from .portfolio import PseudoTimePortfolio
        from .asset import Cash
        
        folio = PseudoTimePortfolio(
            principal=Cash(self.__class__.CURRENCY, cash_quantity),
            cash=Cash(self.__class__.CURRENCY, cash_quantity),
            items=[],
            market=self, 
            **kwargs,
        )
        return folio
        
    def time_sequence(self, offset):
        offset_list = list(offset)
        assert len(offset_list) >= 1, "Expected at least one element"
        assert sorted(offset_list) == offset_list, "Offset must be sorted in ascending order"
        first = offset_list[0]
        assert isinstance(first, timedelta), f"Expected timedelta, got {type(first)} {first}"
        return offset_list
        
    def get_bid_price(self, symbol, time=None):
        return self.get_price(symbol, time=time)
    
    def get_ask_price(self, symbol, time=None):
        return self.get_price(symbol, time=time)
    
    def get_shares(self, price, shares=None, amount=None):
        assert isinstance(price, int) or isinstance(price, float), f"Expected price as int or float, got {type(price)} {price}"
        if shares is not None:
            assert amount is None, "Expected exactly one of 'shares' and 'amount'"
            assert isinstance(shares, int) or isinstance(shares, float), f"Expected shares as int or float, got {type(shares)} {shares}"
        else:
            assert isinstance(amount, float), f"Expected amount to be float, got {type(amount)} {amount}"
            shares = float(amount) / float(price)
        return shares
        
    def market_bid(self, symbol, shares=None, amount=None, time=None):
        from .asset import Cash, PseudoTimeItem
        
        price = self.get_bid_price(symbol, time=time)
        shares = self.get_shares(price, shares=shares, amount=amount)
        item = PseudoTimeItem(name=symbol, quantity=shares, unit_cost=price, market=self)
        cash = Cash(name=self.__class__.CURRENCY, quantity=shares*price)
        return item, cash
    
    def market_ask(self, symbol, shares=None, amount=None, time=None):
        from .asset import Cash, PseudoTimeItem
        
        price = self.get_ask_price(symbol, time=time)
        shares = self.get_shares(price, shares=shares, amount=amount)
        item = PseudoTimeItem(name=symbol, quantity=shares, unit_cost=price, market=self)
        cash = Cash(name=self.__class__.CURRENCY, quantity=shares*price)
        return item, cash
        

class RealTimeMarket(Loggable):
    """
    Base class for real-time markets.
    """
    @abstractmethod
    def market_buy(self, symbol, shares=None, amount=None, **kwargs):
        pass
            
    @abstractmethod
    def market_sell(self, symbol, shares=None, amount=None, **kwargs):
        pass
            
    @abstractmethod
    def get_price(self, symbol):
        pass
    
    @abstractmethod
    def get_bid_price(self, symbol, **kwargs):
        pass
    
    @abstractmethod
    def get_ask_price(self, symbol, **kwargs):
        pass
    


"""
Base class for sending notifications.

Notifications can be for transaction bookkeeping, analytics or interactive decision making.
"""
from tr4d3r.core import Loggable
from abc import abstractmethod
import telegram
from telegram.ext import ExtBot, Updater, MessageHandler, Filters, Defaults
import time
import pytz
import wrappy


class Chat(Loggable):
    def __init__(self, bot_config, client_config):
        self.init_bot(bot_config)
        assert hasattr(self, 'bot'), f"Expecting bot attribute"
        self.init_client(client_config)
        assert hasattr(self, 'client'), f"Expecting client attribute"
    
    @abstractmethod
    def init_bot(self, bot_config):
        pass
    
    @abstractmethod
    def init_client(self, client_config):
        pass
    
    @abstractmethod
    def ping(self, message, **kwargs):
        pass
    

class TelegramChat(Chat):
    """
    Chat based on python-telegram-bot.
    """
    PARSE_MODE = telegram.ParseMode.MARKDOWN_V2
    TIME_ZONE = 'Etc/GMT'
    
    def init_bot(self, bot_config):
        token = bot_config['token']
        defaults = Defaults(
            parse_mode=self.__class__.PARSE_MODE,
            tzinfo=pytz.timezone(self.__class__.TIME_ZONE),
        )
        self.bot = ExtBot(token=token, defaults=defaults)
        self.updater = Updater(token=token, use_context=True, defaults=defaults)
        self._good(f"Created bot {self.bot.get_me()}")
        
    def init_client(self, client_config):
        self.client = client_config['chat_id']
        self._good(f"Created client {self.client}")
    
    def ping(self, text, **kwargs):
        self.bot.send_message(
            text=text,
            chat_id=self.client,
            **kwargs,
        )
        
    def attach(self, document, **kwargs):
        self.bot.send_document(
            document=document,
            chat_id=self.client,
            **kwargs,
        )
        
    @wrappy.todo("Consider the implementation. Should this function exist? Should it be scheduled update fetch and callback instead? Should we use Updater or the current logic?")
    def catch(self, check_period=5):
        """
        Wait for an update from the client and return it.
        """
        updates = []
        while not updates:
            updates = self.bot.get_updates()
            update_dictl = [_u.to_dict() for _u in updates]
            valid_idx = [i for i, _d in enumerate(update_dictl) if _d['chat']['id'] == self.client]
            if valid_updates:
                last_idx = valid_idx[-1]
                last_update = updates[last_idx]
            else:
                time.sleep(check_period)
        return update


from tr4d3r.core import Loggable
from tr4d3r.core.asset import Cash, Item, PseudoTimeItem
from tr4d3r.core.mechanism import NamedCallback, Tentative
from tr4d3r.utils.misc import autostamp, datetime_to_string, string_to_datetime, utcnow
from pprint import pformat
from datetime import datetime
from abc import abstractmethod
import wrappy
import json


class Portfolio(Loggable):
    """
    Base class for portfolios.
    """
    
    DEFAULT_CASH_CLASS = Cash
    DEFAULT_ITEM_CLASS = Item
    DUMMY_ITEM = Item('NULL', quantity=0.0, unit_cost=0.0)
    DEFAULT_GAP_SECONDS = 86400
    
    def __init__(self, principal, cash, items, **kwargs):
        # initialize holdings and other stats
        assert isinstance(principal, Cash)
        assert isinstance(cash, Cash)
        self.principal = principal
        self.cash = cash
        self.item_dict = dict()
        for _item in items:
            assert isinstance(_item, Item)
            self.add_item(_item)
        self.last_tick_time = kwargs.get('last_tick_time', None)
        
        # initialize support parameters
        self.verbose = kwargs.get('verbose', None)
        
        # initialize portfolio management data structure
        self._callbacks = []
       
    def __repr__(self):
        return pformat(self.to_dict())
    
    def __getitem__(self, symbol):
        return self.item_dict.get(symbol, self.__class__.DUMMY_ITEM)
    
    def to_dict(self):
        data_dict = {
            "class": self.__class__.__name__,
            "last_tick_time": self.last_tick_time,
            "principal": self.principal.to_dict(),
            "cash": self.cash.to_dict(),
            "items": {_item.name: _item.to_dict() for _item in sorted(self.item_dict.values(), key=lambda x: x.name)},
        }
        return data_dict
    
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        class_name = burner_dict.pop("class")
        prin_dict = burner_dict.pop("principal")
        cash_dict = burner_dict.pop("cash")
        items_dict = burner_dict.pop("items")
        
        # sanity check
        assert class_name == cls.__name__, f"Class name mismatch: {class_name} vs. {cls.__name__}"
        
        # constructor
        principal = cls.DEFAULT_CASH_CLASS.from_dict(prin_dict)
        cash = cls.DEFAULT_CASH_CLASS.from_dict(cash_dict)
        items = [cls.DEFAULT_ITEM_CLASS.from_dict(_item_d) for _item_d in items_dict.values()]
        return cls(principal, cash, items, **burner_dict)
       
    def dump_json(self, path):
        with open(path, 'w') as f:
            json.dump(self.to_dict(), f, indent=4)
        return path
            
    @classmethod
    def load_json(cls, path):
        with open(path, 'r') as f:
            folio = cls.from_dict(json.load(f))
        return folio
           
    @abstractmethod
    def worth(self, *args, **kwargs):
        pass
           
    @abstractmethod
    def worth_detail(self, *args, **kwargs):
        pass
    
    def cost(self):
        value = self.principal.worth()
        for _item in self.item_dict.values():
            value += _item.cost()
        return value
           
    def cost_detail(self):
        detail = {
            "total": self.cost(),
            "cash": self.principal.__repr__(),
            "items": [{"name": _item.name, "cost": _item.cost(time)} for _item in self.item_dict.values()],
        }
        return detail
    
    def add_item(self, item):
        """
        Add an item to the portfolio.
        """
        name = item.name
        if name in self.item_dict:
            self.item_dict[name].add(item)
        else:
            self.item_dict[name] = item
    
    def sub_item(self, item):
        """
        Subtract an item from the portfolio.
        """
        name = item.name
        if name in self.item_dict:
            self.item_dict[name].sub(item)
            self.check(name)
        else:
            raise KeyError(f"Item {name} does not exist in the portfolio.")
            
    def check(self, name):
        """
        Remove an item if its quantity is numerically zero.
        """
        if name in self.item_dict:
            quantity = self.item_dict[name].quantity
            if -1e-16 < quantity < 1e-16:
                self.item_dict.pop(name)
            elif quantity <= -1e-16:
                raise ValueError(f"Invalid quantity: {quantity} of {name}")
            else:
                pass
            
    @abstractmethod
    def flush_callbacks(self, *args, **kwargs):
        pass

    @abstractmethod
    def tick(self, *args, **kwargs):
        pass

    @abstractmethod
    def deposit(self, amount, **kwargs):
        pass

    @abstractmethod
    def withdraw(self, amount, **kwargs):
        pass

    @abstractmethod
    def market_buy(self, symbol, shares=None, amount=None, **kwargs):
        pass

    @abstractmethod
    def market_sell(self, symbol, shares=None, amount=None, **kwargs):
        pass

    
class PseudoTimePortfolio(Portfolio):
    """
    Portfolio specifically for pseudo-time simulation.
    """
    
    DEFAULT_CASH_CLASS = Cash
    DEFAULT_ITEM_CLASS = PseudoTimeItem
    DUMMY_ITEM = PseudoTimeItem('NULL', quantity=0.0, unit_cost=0.0, market=None)
    DEFAULT_GAP_SECONDS = 86400
    
    def __init__(self, principal, cash, items, market, **kwargs):
        super().__init__(principal, cash, items, **kwargs)
        if isinstance(self.last_tick_time, str):
            self.last_tick_time = string_to_datetime(self.last_tick_time)
        self.market = market
            
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        class_name = burner_dict.pop("class")
        prin_dict = burner_dict.pop("principal")
        cash_dict = burner_dict.pop("cash")
        items_dict = burner_dict.pop("items")
        market = burner_dict.pop("market")
        
        # sanity check
        assert class_name == cls.__name__, f"Class name mismatch: {class_name} vs. {cls.__name__}"
        
        # constructor
        principal = cls.DEFAULT_CASH_CLASS.from_dict(prin_dict)
        cash = cls.DEFAULT_CASH_CLASS.from_dict(cash_dict)
        items = [cls.DEFAULT_ITEM_CLASS.from_dict(_item_d) for _item_d in items_dict.values()]
        return cls(principal, cash, items, market, **burner_dict)
       
    def worth(self, time):
        value = self.cash.worth()
        for _item in self.item_dict.values():
            value += _item.worth(time)
        return value
    
    def worth_detail(self, time):
        detail = {
            "total": self.worth(time),
            "cash": self.cash.__repr__(),
            "items": [{"name": _item.name, "worth": _item.worth(time)} for _item in self.item_dict.values()],
        }
        return detail
        
    def deposit(self, amount, time):
        self.principal.quantity += amount
        self.cash.quantity += amount
        self._info(f"Cash deposit {amount} @ {time} | balance = {self.cash.quantity}")

    def withdraw(self, amount, time):
        self.principal.quantity -= amount
        self.cash.quantity -= amount
        self._info(f"Cash withdrawal {amount} @ {time} | balance = {self.cash.quantity}")

    def tick(self, time):
        gap_seconds = self.__class__.DEFAULT_GAP_SECONDS if self.last_tick_time is None else (time - self.last_tick_time).total_seconds()
        assert gap_seconds > 0, f"Unexpected timeline: last move @ {self.last_tick_time}, tick @ {time}"
        self.last_tick_time = time
        return gap_seconds
        
    def market_buy(self, symbol, shares=None, amount=None, time=None):
        time = autostamp(time)
        item, cash = self.market.market_ask(symbol, shares=shares, amount=amount, time=time)
        if self.verbose:
            self._good(f"{datetime.fromtimestamp(time)}: B {round(item.quantity, 6)} {item.name} <- {round(cash.quantity, 6)} {cash.name}.")
        self.cash.sub(cash)
        self.add_item(item)
    
    def market_sell(self, symbol, shares=None, amount=None, time=None):
        time = autostamp(time)
        item, cash = self.market.market_bid(symbol, shares=shares, amount=amount, time=time)
        if self.verbose:
            self._good(f"{datetime.fromtimestamp(time)}: S {round(item.quantity, 6)} {item.name} -> {round(cash.quantity, 6)} {cash.name}.")
        self.sub_item(item)
        self.cash.add(cash)
        self.check(item.name)
    
    def limit_buy(self, symbol, shares, limit_price, order_time=None, expire_after=86400):
        
        def condition_f(time):
            price = self.market.get_ask_price(symbol, time=time)
            return bool(price <= limit_price)
        
        def trigger_f(time):
            self.market_buy(symbol, shares=shares, time=time)
            
        condition = NamedCallback(
            condition_f, 
            f"{symbol} P<={round(limit_price, 6)}"
        )
        trigger = NamedCallback(
            trigger_f,
            f"B {round(shares, 6)} * {symbol}"
        )
        callback = Tentative(
            condition=condition,
            trigger=trigger,
            order_time=autostamp(order_time),
            expire_after=86400,
        )
        
        self._callbacks.append(callback)
    
    def limit_sell(self, symbol, shares, limit_price, order_time=None, expire_after=86400):
        
        def condition_f(time):
            price = self.market.get_bid_price(symbol, time=time)
            return bool(price >= limit_price)
        
        def trigger_f(time):
            self.market_sell(symbol, shares=shares, time=time)
            
        condition = NamedCallback(
            condition_f, 
            f"{symbol} P>={round(limit_price, 6)}"
        )
        trigger = NamedCallback(
            trigger_f,
            f"S {round(shares, 6)} * {symbol}"
        )
        callback = Tentative(
            condition=condition,
            trigger=trigger,
            order_time=autostamp(order_time),
            expire_after=86400,
        )
        
        self._callbacks.append(callback)

    def flush_callbacks(self, time):
        callbacks = self._callbacks[:]
        self._callbacks.clear()
        
        for _callback in callbacks:
            _done = _callback(time)
            if not _done:
                self._callbacks.append(_callback)


class RealTimePortfolio(Portfolio):
    """
    Base class for portfolios for real-time trading.
    """
    DEFAULT_GAP_SECONDS = 86400
    
    def __init__(self, principal, cash, items, open_orders, **kwargs):
        super().__init__(principal, cash, items, **kwargs)
        if isinstance(self.last_tick_time, str):
            self.last_tick_time = string_to_datetime(self.last_tick_time)
        # real-time portfolios have open orders that get closed later
        self._open_orders = set(open_orders)
       
    def to_dict(self):
        data_dict = {
            "class": self.__class__.__name__,
            "last_tick_time": datetime_to_string(self.last_tick_time),
            "principal": self.principal.to_dict(),
            "cash": self.cash.to_dict(),
            "items": {_item.name: _item.to_dict() for _item in sorted(self.item_dict.values(), key=lambda x: x.name)},
            "open_orders": list(self.open_orders),
        }
        return data_dict
    
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        class_name = burner_dict.pop("class")
        prin_dict = burner_dict.pop("principal")
        cash_dict = burner_dict.pop("cash")
        items_dict = burner_dict.pop("items")
        open_orders = burner_dict.pop("open_orders")
        
        # sanity check
        assert class_name == cls.__name__, f"Class name mismatch: {class_name} vs. {cls.__name__}"
        
        # constructor
        principal = cls.DEFAULT_CASH_CLASS.from_dict(prin_dict)
        cash = cls.DEFAULT_CASH_CLASS.from_dict(cash_dict)
        items = [cls.DEFAULT_ITEM_CLASS.from_dict(_item_d) for _item_d in items_dict.values()]
        return cls(principal, cash, items, open_orders, **burner_dict)
       
    @property
    def open_orders(self):
        return self._open_orders
    
    @open_orders.setter
    def open_orders(self, orders):
        raise ValueError("It is forbidden to directly set open orders. Please use set methods such as add() or remove().")
        
    @wrappy.todo("No programmatic deposit implemented at the moment.")
    def deposit(self, amount, **kwargs):
        pass

    @wrappy.todo("No programmatic withdrawal implemented at the moment.")
    def withdraw(self, amount, **kwargs):
        pass

    def worth(self):
        value = self.cash.worth()
        for _item in self.item_dict.values():
            value += _item.worth()
        return value
    
    def worth_detail(self):
        detail = {
            "total": self.worth(),
            "cash": self.cash.__repr__(),
            "items": [{"name": _item.name, "worth": _item.worth()} for _item in self.item_dict.values()],
        }
        return detail
       
    @abstractmethod
    def open_order_values(self):
        """
        Compute the values of open orders for each item symbol.
        Buy -> positive
        Sell -> negative
        """
        pass
        
    def tick(self, update):
        now_utc = utcnow()
        gap_seconds = self.__class__.DEFAULT_GAP_SECONDS if self.last_tick_time is None else (now_utc - self.last_tick_time).total_seconds()
        assert gap_seconds > 0, f"Unexpected timeline: last move @ {self.last_tick_time}, tick @ {now_utc}"
        assert isinstance(update, bool), f"Expected boolean value"
        if update:
            self.last_tick_time = now_utc
        return gap_seconds
        
    def market_buy(self, symbol, shares=None, amount=None, **kwargs):
        order = self.market.market_buy(symbol, shares=shares, amount=amount, **kwargs)
        if order is not None:
            self.open_orders.add(order['id'])
        return order

    def market_sell(self, symbol, shares=None, amount=None, **kwargs):
        order = self.market.market_sell(symbol, shares=shares, amount=amount, **kwargs)
        if order is not None:
            self.open_orders.add(order['id'])
        return order
    
    @abstractmethod
    def refresh_open_orders(self):
        """
        Refresh the state of all open orders.
        """
        pass
    
    def flush_callbacks(self):
        callbacks = self._callbacks[:]
        self._callbacks.clear()
        
        for _callback in callbacks:
            _done = _callback()
            if not _done:
                self._callbacks.append(_callback)



"""
Mechanism helpers.
"""
from tr4d3r.core import Loggable
from tr4d3r.utils.misc import autostamp
from datetime import datetime


class NamedCallback(Loggable):
    def __init__(self, f, name):
        self._f = f
        self.name = name
        
    def __call__(self, *args, **kwargs):
        return self._f(*args, **kwargs)
    
    def __repr__(self):
        return self.name
    
    
class Tentative(Loggable):
    def __init__(self, condition, trigger, order_time=None, expire_after=86400, verbose=False):
        """
        Trigger a function conditionally at most once with expiration.
        """
        # sanity check
        assert not condition(order_time), f"Condition is not expected to be met immediately"
        
        # initialization
        self.order_time = autostamp(order_time)
        self.expire_time = order_time + expire_after
        self.state_dict = {"expired": False, "triggered": False, "latest": order_time}
        self.condition = condition
        self.trigger = trigger
        self.verbose = bool(verbose)
    
    def __repr__(self):
        status = "expired" if self.state_dict["expired"] else "pending"
        status = "complete" if self.state_dict["triggered"] else "pending"
        read_time = datetime.fromtimestamp(self.state_dict['latest'])
        return f"{self.condition.name} -> {self.trigger.name} | {status} @ {read_time}"
        
    def __call__(self, time):
        """
        Return whether this tentative should be marked as complete.
        """
        # base case: expired or already triggered
        if self.state_dict["expired"] or self.state_dict["triggered"]:
            raise ValueError(f"Calling invalid tentative {self}.")
            #return True
        
        # sanity check: timeline should be consistent
        time = autostamp(time)
        last_time = self.state_dict["latest"]
        assert time >= last_time, f"Illegal limit order state: old time {last_time}, new time {time}"
        
        # exit: order expires
        if time >= self.expire_time:
            self.state_dict["expired"] = True
            self.state_dict["latest"] = time
            if self.verbose:
                self._info(f"{self}")
            return True
        
        # exit: condition is met
        if self.condition(time):
            self.trigger(time)
            self.state_dict["triggered"] = True
            self.state_dict["latest"] = time
            if self.verbose:
                self._info(f"{self}")
            return True
        
        # keep going: order still stands
        self.state_dict["latest"] = time
        return False


"""
Simulation classes.
"""
from tr4d3r.core import Loggable
from tqdm import tqdm
import pandas as pd


class Simulation(Loggable):
    """
    Simulation with pseuto-time objects.
    """
    def __init__(self, market_cls, *args, **kwargs):
        self.init_kwargs = kwargs
        self.market = market_cls(*args, **kwargs)
        
    def run(self, manager, folio, time_sequence, timed_callbacks=None):
        """
        TODO: cache sequences and load when a new sequence is a contiguous subset with the same start.
        """
        timed_callbacks = timed_callbacks or {}
        
        data_entries = []
        for _time in tqdm(time_sequence, desc="Simulation step"):
            # find any callback and trigger it
            if _time in timed_callbacks.keys():
                _callback = timed_callbacks[_time]
                _callback(_time)
            # regular read/write subroutine
            data_entries.append(manager.tick_read(folio, _time))
            manager.tick_write(folio, _time)
        
        df = pd.DataFrame(data_entries)
        info = {
            'portfolio': folio,
            'market': self.market,
            'df': df,
        }
        return info


"""
Asset components.
"""
from tr4d3r.core import Loggable
from tr4d3r.utils.misc import autostamp
from pprint import pformat
from datetime import datetime
from abc import abstractmethod


class Asset(Loggable):
    """
    Anything of value.
    """
    def __init__(self, name, quantity):
        self._name = name
        self._quantity = quantity
    
    @property
    def name(self):
        return self._name
    
    @property
    def quantity(self):
        return self._quantity
    
    @quantity.setter
    def quantity(self, value):
        assert value >= 0.0, f"Expecting non-negative value, got {value}"
        self._quantity = value
    
    def check(self, other):
        assert isinstance(other, self.__class__)
        assert self.name == other.name, f"Conflicting asset name: {self.name} vs. {other.name}"
    
    @abstractmethod
    def worth(self):
        pass
    
    @property
    def cost(self):
        pass
    
    
class Cash(Asset):
    """
    Whatever is used for purchases and for measuring value.
    """
    def __init__(self, name, quantity):
        super().__init__(name, quantity)
    
    def worth(self):
        return self.quantity
        
    def __repr__(self):
        return f"{self.name}: {round(self.quantity, 6)} units"
    
    def to_dict(self):
        data_dict = {
            "class": self.__class__.__name__,
            "name": self.name,
            "quantity": self.quantity
        }
        return data_dict
    
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        class_name = burner_dict.pop("class")
        name = burner_dict.pop("name")
        quantity = burner_dict.pop("quantity")
        
        # sanity check
        assert class_name == cls.__name__, f"Class name mismatch: {class_name} vs. {cls.__name__}"
        assert len(burner_dict) == 0, f"Expecting no remaining attributes, got {burner_dict.keys}"
        return cls(name, quantity)
        
    def add(self, other):
        if other is None:
            return
        self.check(other)
        self.quantity += other.quantity
        other.quantity = 0
    
    def sub(self, other):
        if other is None:
            return
        self.check(other)
        self.quantity -= other.quantity
        other.quantity = 0
        
        
class Item(Asset):
    """
    Whatever is purchased and has variable value.
    """
    def __init__(self, name, quantity, unit_cost):
        super().__init__(name, quantity)
        self._unit_cost = unit_cost
    
    def __repr__(self):
        return f"{self.name}: {round(self.quantity, 6)} units, cost {round(self.unit_cost, 6)} per unit"
    
    def to_dict(self):
        data = {
            "class": self.__class__.__name__,
            "name": self.name, 
            "quantity": self.quantity, 
            "unit_cost": self.unit_cost}
        return data
    
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        class_name = burner_dict.pop("class")
        name = burner_dict.pop("name")
        quantity = burner_dict.pop("quantity")
        unit_cost = burner_dict.pop("unit_cost")
        
        # sanity check
        assert class_name == cls.__name__, f"Class name mismatch: {class_name} vs. {cls.__name__}"
        assert len(burner_dict) == 0, f"Expecting no remaining attributes, got {burner_dict.keys()}"
        return cls(name, quantity, unit_cost)
    
    @property
    def unit_cost(self):
        return self._unit_cost
    
    @unit_cost.setter
    def unit_cost(self, value):
        assert value >= 0.0, f"Expecting non-negative value, got {value}"
        self._unit_cost = value
        
    def cost(self):
        return self.unit_cost * self.quantity
    
    def worth(self, *args, **kwargs):
        return self.unit_price(*args, **kwargs) * self.quantity
    
    def bid_worth(self, *args, **kwargs):
        return self.bid_price(*args, **kwargs) * self.quantity
        
    def ask_worth(self, *args, **kwargs):
        return self.ask_price(*args, **kwargs) * self.quantity
    
    @abstractmethod
    def unit_price(self, *args, **kwargs):
        pass
    
    @abstractmethod
    def bid_price(self, *args, **kwargs):
        pass
    
    @abstractmethod
    def ask_price(self, *args, **kwargs):
        pass
    
    def add(self, other):
        if other is None:
            return
        self.check(other)
        total_cost = self.cost() + other.cost()
        total_quantity = self.quantity + other.quantity
        average_unit_cost = total_cost / total_quantity
        self.quantity = total_quantity
        self.unit_cost = average_unit_cost
        other.quantity = 0
    
    def sub(self, other):
        if other is None:
            return
        self.check(other)
        self.quantity -= other.quantity
        other.quantity = 0
    

class PseudoTimeItem(Item):
    """
    Item for pseudo-time simulation where the market can be a variable.
    """
    from .market import PseudoTimeMarket
    COMPATIBLE_MARKETS = {
        PseudoTimeMarket,
    }
    
    def __init__(self, name, quantity, unit_cost, market):
        super().__init__(name, quantity, unit_cost)
        
        # sanity check: market must be None or from compatible classes
        valid_market = bool(market is None)
        for _cls in self.__class__.COMPATIBLE_MARKETS:
            if isinstance(market, _cls):
                valid_market = True
        assert valid_market, f"Unexpected market type: {market.__class__}"
        self._market = market
    
    def to_dict(self):
        data = {"name": self.name, "quantity": self.quantity, "unit_cost": self.unit_cost, "market": self._market}
        return data
    
    @classmethod
    def from_dict(cls, data_dict):
        burner_dict = data_dict.copy()
        name = burner_dict.pop("name")
        quantity = burner_dict.pop("quantity")
        unit_cost = burner_dict.pop("unit_cost")
        market = burner_dict.pop("market")
        assert len(burner_dict) == 0, f"Expecting no remaining attributes, got {burner_dict.keys()}"
        return cls(name, quantity, unit_cost, market)
    
    def unit_price(self, time):
        if self._market is None:
            return 0.0
        return self._market.get_price(self.name, time=time)

    def bid_price(self, time):
        if self._market is None:
            return 0.0
        return self._market.get_bid_price(self.name, time=time)

    def ask_price(self, time):
        if self._market is None:
            return 0.0
        return self._market.get_ask_price(self.name, time=time)




import setuptools
import os


def get_description():
    if os.path.isfile("README.md"):
        with open("README.md", "r") as fh:
            desc = fh.read()
    else:
        desc = ""
    return desc


setuptools.setup(
    name="tr4d3r",
    version="0.1.2",
    description="A rebalancing trade bot library. A good one.",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    author="Pavel",
    author_email="pavelhurwicz@gmail.com",
    url="https://github.com/phurwicz/tr4d3r",
    packages=setuptools.find_packages(),
    install_requires=[
        # trading backend
        "robin_stocks",
        # messaging backend
        "python-telegram-bot",
        # data handling
        "numpy>=1.14",
        "pandas>=1.1.4",
        # utilities
        "deprecated",
        "dill",
        "rich",
        "tqdm",
        "wasabi",
        "wrappy",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Random-number utility functions.
"""
import numpy as np


def exponential_random_walk(rng, initial, steps, **kwargs):
    """
    An expontential random walk from a scalar or vector of initial values.
    """
    # determine dimensionality based on initial values and number of steps
    shape_initial = np.asarray(initial).shape
    assert len(shape_initial) <= 1, f"Expected scalar or 1-D array, got {initial}"
    shape_steps = (steps, *shape_initial)
    
    # sample individual steps and cumsum for random walk
    log_initial = np.log(initial)
    log_initial_broadcast = log_initial[np.newaxis] if len(shape_initial) == 0 else log_initial[np.newaxis, :]
    log_increments = rng.normal(size=shape_steps, **kwargs)
    log_walk = np.concatenate(
        (log_initial_broadcast, log_increments),
        axis=0,
    ).cumsum(axis=0)
    final_walk = np.exp(log_walk)
    return final_walk
